/*
 * Copyright (c) 2005 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
 * spnegoKrb.h - Kerberos routines for SPNEGO tool
 *
 * Created July 7 2003 by dmitch 
 */
 
#ifndef	_SPNEGO_KRB_H_
#define _SPNEGO_KRB_H_

//#include "spnegoTool.h"
#include <Kerberos/krb5.h>		/* just for krb5_error_code */

#ifdef __cplusplus
extern "C" {
#endif

krb5_error_code GetSvcTicketForHost(const char *inHostname,
									const char *inServiceType,
									unsigned *outTicketLen,
									unsigned char **outTicket);
	
/*
 * Obtain service ticket for specified principal. 
 * Result is malloc'd; caller must free.
 */
krb5_error_code krbGetSvcTkt(
	const char *principal,
	unsigned char **tkt,		// mallocd and RETURNED
	unsigned *tktLen,		// RETURNED
	int *isLoggedIn);		// RETURNED: nonzero means the user
                                        // has a valid TGT

#ifdef __cplusplus
}
#endif

#endif	/* _SPNEGO_KRB_H_ */
